/*
    Copyright 2016 Gleb Popov <6yearold@gmail.com>

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:
    1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
    3. Neither the name of the copyright holder nor the names of its contributors may
    be used to endorse or promote products derived from this software without specific
    prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
    OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#pragma once

#include <QObject>
#include <QDBusObjectPath>
#include <QDebug>
#include <QUrl>

#include "geomprober.h"
#include "blockfilesystem.h"
#include "blockpartition.h"
#include "blockparttable.h"

class Block : public QObject
{
    Q_OBJECT
public:
    Block(QObject* parent = 0);

    BlockFilesystem* bFilesystem;
    BlockPartition* bPartition;
    BlockPartTable* bPartTable;

    QString dev;
    QString description;
    QString identifier;
    bool geomCreated;
    QDBusObjectPath dbusPath;
    QStringList labels;

    Q_PROPERTY(QByteArray Device READ device)
    QByteArray device() const
    {
        return dev.toLatin1();
    }

    Q_PROPERTY(QByteArray PreferredDevice READ preferredDevice)
    QByteArray preferredDevice() const
    {
        if(labels.empty())
            return device();
        return (QStringLiteral("/dev/") + labels[0]).toLatin1();
    }

    Q_PROPERTY(QString IdLabel READ idLabel)
    QString idLabel() const
    {
        return labels.empty()
            ? QString()
            : QUrl::fromPercentEncoding(labels[0].mid(labels[0].lastIndexOf('/') + 1).toLatin1());
    }

    Q_PROPERTY(QStringList Symlinks READ symlinks)
    QStringList symlinks()
    {
        QStringList r;
        foreach(auto s, labels)
            r << QStringLiteral("/dev/") + s;
        return r;
    }

    Q_PROPERTY(QString Id READ id)
    QString id() const
    {
        if(bPartition)
        {
            auto tableType = bPartition->partTableBlock->bPartTable->tableType();
            QChar partitionSymbol = tableType == QStringLiteral("gpt")
                                  ? 'p'
                                  : 's';
            return bPartition->partTableBlock->id() + "_" + dev.mid(dev.lastIndexOf(partitionSymbol));
        }
        else
            return description + "_" + identifier;
    }

    QString idType;
    Q_PROPERTY(QString IdType MEMBER idType)

    QString idUsage;
    Q_PROPERTY(QString IdUsage MEMBER idUsage)

    qulonglong size;
    Q_PROPERTY(qulonglong Size READ blockSize)
    qulonglong blockSize() const
    {
        return bPartition ? bPartition->size : size;
    }

    // =======================================
    // Filesystem

    Q_PROPERTY(QByteArrayList MountPoints READ mountPoints);
    QByteArrayList mountPoints() const
    {
        return bFilesystem ? bFilesystem->mountPoints : QByteArrayList();
    }

    // =======================================
    // PartitionTable

    QHash<QString, Part> parts;

    Q_PROPERTY(QString PartitionTable_Type READ partitionTableType)
    QString partitionTableType()
    {
        if(!bPartTable)
            return QString();

        return bPartTable->tableType();
    }

    // =======================================
    // Partition

    Q_PROPERTY(QString Partition_Name READ partitionName)
    QString partitionName() const
    {
        if(!bPartition)
            return QString();

        return bPartition->name();
    }

    Q_PROPERTY(uint Number READ number)
    uint number() const
    {
        if(!bPartition)
            return 0;

        return bPartition->number;
    }

    Q_PROPERTY(qulonglong Offset READ offset)
    qulonglong offset() const
    {
        if(!bPartition)
            return 0;

        return bPartition->offset;
    }

    Q_PROPERTY(qulonglong Partition_Size READ partitionSize)
    qulonglong partitionSize() const
    {
        if(!bPartition)
            return 0;

        return bPartition->size;
    }

    Q_PROPERTY(QString Partition_Type READ partitionType)
    QString partitionType() const
    {
        if(!bPartition)
            return 0;

        return bPartition->partitionType;
    }

    Q_PROPERTY(QDBusObjectPath Table READ table)
    QDBusObjectPath table() const
    {
        if(!bPartition)
            return QDBusObjectPath("/");

        return bPartition->partTableBlock->dbusPath;
    }

    // =======================================
    // Unimplemented Block stuff

    Q_PROPERTY(QDBusObjectPath CryptoBackingDevice READ cryptoBackingDevice)
    QDBusObjectPath cryptoBackingDevice() const
    {
        return QDBusObjectPath("/");
    }

    Q_PROPERTY(QDBusObjectPath Drive READ drive)
    QDBusObjectPath drive() const
    {
        return QDBusObjectPath("/");
    }

    Q_PROPERTY(QDBusObjectPath MDRaid READ mDRaid)
    QDBusObjectPath mDRaid() const
    {
        return QDBusObjectPath("/");
    }

    Q_PROPERTY(QDBusObjectPath MDRaidMember READ mDRaidMember)
    QDBusObjectPath mDRaidMember() const
    {
        return QDBusObjectPath("/");
    }

public slots:
    QString Mount(const QVariantMap &options)
    {
        if(!bFilesystem)
            return QString();

        return bFilesystem->Mount(options);
    }
    void Unmount(const QVariantMap &options)
    {
        if(!bFilesystem)
            return;

        return bFilesystem->Unmount(options);
    }
};
