/*
    Copyright 2016 Gleb Popov <6yearold@gmail.com>

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:
    1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
    3. Neither the name of the copyright holder nor the names of its contributors may
    be used to endorse or promote products derived from this software without specific
    prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
    OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "blockfilesystem.h"
#include "block.h"

#include <QDBusConnection>
#include <QDBusMessage>
#include <QProcess>
#include <QDir>

QString createMountPoint(QString id, int suff = 0)
{
    auto mp = suff == 0 ? QStringLiteral("/media/") + id
                        : QStringLiteral("/media/") + id + QString::number(suff);
    QDir mpDir(mp);

    if(mpDir.exists())
        return createMountPoint(id, ++suff);

    mpDir.mkpath(".");

    return mp;
}

void removeMountPoint(QString mp, bool checkIfEmpty = false)
{
    QDir mpDir(mp);

    if(checkIfEmpty && mpDir.entryList().count() > 2) // '.' and '..' also counts
        return;

    auto dirName = mpDir.dirName();
    mpDir.cdUp();
    mpDir.rmdir(dirName);
}

QString BlockFilesystem::Mount(const QVariantMap& options)
{
    QProcess mount;
    QString mountPoint = createMountPoint(parentBlock()->id().replace(' ', '_'));
    QStringList args;

    auto mountProg = QStringLiteral("/sbin/mount");
    if(filesystem == "msdosfs")
        mountProg = QStringLiteral("/sbin/mount_msdosfs");
    if(filesystem == "ntfs")
    {
        // TODO: when bsdisks is started by dbus, calling ntfs-3g yields
        // error without any error message. I wasn't able to find out why
        // To reproduce run
        // sudo -u messagebus -- /usr/local/libexec/dbus-daemon-launch-helper org.freedesktop.UDisks2
        // and then try to mount any ntfs flash disk with, say, qdbusviewer or solid-hardware
        bool workaround = true;
        if(!workaround)
            mountProg = QStringLiteral("ntfs-3g");
        else
        {
            mountProg = QStringLiteral("env");
            args << QStringLiteral("PATH=/sbin:/usr/sbin:/usr/local/bin")
                 << QStringLiteral("ntfs-3g");
        }
    }

    args << parentBlock()->dev << mountPoint;

    mount.setProgram(mountProg);
    mount.setArguments(args);

    mount.start();
    mount.waitForFinished(-1);

    if(mount.exitCode())
    {
        qDebug() << "mount error: " << mount.readAllStandardError();
        removeMountPoint(mountPoint);
        return QString();
    }

    mountPoints << mountPoint.toUtf8();
    
    QVariantMap props;
    props.insert(QStringLiteral("MountPoints"), QVariant::fromValue(mountPoints));
    
    QDBusMessage signal = QDBusMessage::createSignal(
                                parentBlock()->dbusPath.path(),
                                QStringLiteral("org.freedesktop.DBus.Properties"),
                                QStringLiteral("PropertiesChanged"));
    
    signal << QStringLiteral("org.freedesktop.UDisks2.Filesystem")
           << props
           << QStringList();
    QDBusConnection::systemBus().send(signal);
    
    return mountPoint;
}

void BlockFilesystem::Unmount(const QVariantMap& options)
{
    foreach(auto mp, mountPoints)
    {
        QProcess umount;

        QStringList args;
        for(auto it = options.cbegin(); it != options.cend(); it++)
        {
            args << it.key();
            if(!it.value().isNull())
                args << it.value().toString();
        }
        args << QString::fromUtf8(mp);

        umount.start(QStringLiteral("/sbin/umount"), args);
        umount.waitForFinished(-1);

        if(umount.exitCode() == 0)
            removeMountPoint(mp, /*checkIfEmpty = */ true);
    }

    mountPoints.clear();
}

BlockFilesystem::~BlockFilesystem()
{
    QVariantMap args;
    args["-f"] = QVariant();
    Unmount(args);
}

Block* BlockFilesystem::parentBlock()
{
    return qobject_cast<Block*>(parent());
}

BlockFilesystem::BlockFilesystem(Block* parent)
    : QObject(parent)
{
}
