/*
    Copyright 2016 Gleb Popov <6yearold@gmail.com>

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:
    1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
    3. Neither the name of the copyright holder nor the names of its contributors may
    be used to endorse or promote products derived from this software without specific
    prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
    OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <QSet>

#include "manager.h"
#include "adaptors.h"
#include "block.h"
#include "drive.h"
#include "geomprober.h"
#include "filesystemprober.h"
#include "devdthread.h"

const QString UDisksService = QStringLiteral("/org/freedesktop/UDisks2/");
const QString UDisksBlockDevices = QStringLiteral("/org/freedesktop/UDisks2/block_devices/");
const QString UDisksDrives = QStringLiteral("/org/freedesktop/UDisks2/drives/");

class ObjectManager : public QObject
{
    Q_OBJECT
public slots:
    DBUSManagerStruct GetManagedObjects()
    {
        DBUSManagerStruct r;
        return r;
    }

    void geomAdded(QString dev)
    {
        // don't do anything with stuff like diskid/blabla
        if(dev.contains('/'))
            return;

        auto b = m_blockObjects.value(dev);
        if(!b)
        {
            b = new Block;
            b->geomCreated = true;
            b->dbusPath = QDBusObjectPath(UDisksBlockDevices + dev);
            m_blockObjects.insert(dev, b);
            return;
        }

        startFilesystemProbe(b);
        startGeomProbe(b);

        if(b->bPartition)
            updatePartitions(b->bPartition->partTableBlock);
    }

    void filesystemAdded(Block* b, QString fs)
    {
        Q_ASSERT(b->bFilesystem == nullptr);

        auto bfs = new BlockFilesystem(b);
        bfs->filesystem = fs;

        b->bFilesystem = bfs;
        b->idUsage = QStringLiteral("filesystem");

        if(fs == QStringLiteral("msdosfs"))
        {
            b->idType = QStringLiteral("vfat");
        }

        new FilesystemAdaptor(b);
        addInterface(b->dbusPath, "org.freedesktop.UDisks2.Filesystem");
    }

    void removeBlock(QString dev)
    {
        if(m_blockObjects.contains(dev))
        {
            QString devPath = UDisksBlockDevices + dev;
            qDebug() << "Unregistering " + devPath;
            QDBusConnection::systemBus().unregisterObject(devPath);
            delete m_blockObjects.take(dev);
        }
    }

    void addBlock(QString dev)
    {
        // don't do anything with stuff like diskid/blabla
        if(dev.contains('/'))
            return;

        if(dev.startsWith("da"))
        {
            auto b = m_blockObjects.value(dev);
            if(!b)
            {
                b = new Block;
                m_blockObjects.insert(dev, b);
            }

            QString devPath = UDisksBlockDevices + dev;
            b->dev = ("/dev/" + dev).toLatin1();
            b->dbusPath = QDBusObjectPath(devPath);

            new BlockAdaptor(b);

            qDebug() << "Registering " + devPath;
            QDBusConnection::systemBus().registerObject(devPath, b);

            addInterface(b->dbusPath, "org.freedesktop.UDisks2.Block");

            if(b->geomCreated)
            {
                startFilesystemProbe(b);
                startGeomProbe(b);
            }
        }
    }
    void removeDrive(QString dev)
    {
        QString devPath = UDisksDrives + dev;
        qDebug() << "Unregistering " + devPath;
        QDBusConnection::systemBus().unregisterObject(devPath);
        delete m_driveObjects.take(dev);
    }
    void addDrive(QString dev)
    {
        if(dev.startsWith("umass"))
        {
            auto d = m_driveObjects.value(dev);
            if(!d)
            {
                d = new Drive;
                m_driveObjects.insert(dev, d);
            }

            QString devPath = UDisksDrives + dev;
            d->dbusPath = QDBusObjectPath(devPath);
            new DriveAdaptor(d);

            qDebug() << "Registering " + devPath;
            QDBusConnection::systemBus().registerObject(devPath, d);

            addInterface(d->dbusPath, "org.freedesktop.UDisks2.Drive");
        }
    }

    void onInitialProbe(QSet<QString> foundDevs)
    {
        foreach(auto dev, foundDevs)
            geomAdded(dev);
    }
signals:
    void InterfacesAdded(const QDBusObjectPath &object_path, const QVariantMapMap &interfaces_and_properties);

private:
    void startFilesystemProbe(Block* b)
    {
        auto prober = new FilesystemProber(b->dev);

        QObject::connect(prober, &FilesystemProber::finished, this,
            [this, b](QString fs)
            {
                if(!fs.isEmpty())
                    filesystemAdded(b, fs);
            });

        QThreadPool::globalInstance()->start(prober);
    }
    void startGeomProbe(Block* b)
    {
        QFileInfo info(b->dev);
        auto prober = new GeomProber(info.fileName());

        QObject::connect(prober, &GeomProber::gotLabels, this,
            [this, b](QStringList labels)
            {
                b->labels = labels;
            });
        QObject::connect(prober, &GeomProber::gotDisk, this,
            [this, b](DiskInfo di)
            {
                b->size = di.mediaSize();
                b->description = di.descr();
                b->identifier = di.ident();
            });
        QObject::connect(prober, &GeomProber::gotPart, this,
            [this, b](PartInfo pi)
            {
                Q_ASSERT(!b->bPartTable);

                b->bPartTable = new BlockPartTable(b);
                b->bPartTable->setTableType(pi.scheme());
                b->parts = pi.partitions();

                new PartitionTableAdaptor(b);
                addInterface(b->dbusPath, QStringLiteral("org.freedesktop.UDisks2.PartitionTable"));

                updatePartitions(b);
            });

        QThreadPool::globalInstance()->start(prober);
    }

    void updatePartitions(Block* table)
    {
        for(auto it = m_blockObjects.constBegin(); it != m_blockObjects.cend(); it++)
        {
            QFileInfo info(table->dev);

            if(!it.key().startsWith(info.fileName()))
                continue;
            if(it.key() == info.fileName())
                continue;

            Block* b = it.value();

            auto bp = new BlockPartition(b);
            auto pi = table->parts[it.key()];

            // TODO: find out why there are "da0s1" and "/dev/da0s1" keys
            bp->partTableBlock = table;
            bp->partitionType = pi.type();
            bp->number = pi.index();
            bp->offset = pi.offset();
            bp->size = pi.length();

            b->bPartition = bp;

            new PartitionAdaptor(b);
            addInterface(b->dbusPath, "org.freedesktop.UDisks2.Partition");
        }
        table->parts.clear();
    }

    void addInterface(QDBusObjectPath path, QString iface)
    {
        QVariantMapMap interfaces;
        interfaces[iface] = QVariantMap();
        emit InterfacesAdded(path, interfaces);
    }

    QHash<QString, Block*> m_blockObjects;
    QHash<QString, Drive*> m_driveObjects;
};

ObjectManager manager;

void initialGeomProbe()
{
    GeomProber prober;
    QSet<QString> foundDevs;

    QObject::connect(&prober, &GeomProber::gotDisk,
        [&foundDevs](DiskInfo di)
        {
            manager.addBlock(di.devName());
            foundDevs.insert(di.devName());
        });
    QObject::connect(&prober, &GeomProber::gotPart,
        [&foundDevs](PartInfo pi)
        {
            foreach(auto devName, pi.partitions().keys())
            {
                manager.addBlock(devName);
                foundDevs.insert(devName);
            }
        });

    prober.run();

    QMetaObject::invokeMethod(&manager, "onInitialProbe", Qt::QueuedConnection, Q_ARG(QSet<QString>, foundDevs));
}

int main(int argc, char** argv)
{
    qRegisterMetaType<QVariantMapMap>();
    qRegisterMetaType<DBUSManagerStruct>();
    qRegisterMetaType<QSet<QString>>();
    qRegisterMetaType<DiskInfo>();
    qRegisterMetaType<Part>();
    qRegisterMetaType<PartInfo>();

    qDBusRegisterMetaType<QByteArrayList>();
    qDBusRegisterMetaType<QVariantMapMap>();
    qDBusRegisterMetaType<DBUSManagerStruct>();

    QCoreApplication::setSetuidAllowed(true);
    QCoreApplication app(argc, argv);

    if(!QDBusConnection::systemBus().registerService("org.freedesktop.UDisks2"))
    {
        qCritical() << "Could not register UDisks2 service";
        return -1;
    }

    new ObjectManagerAdaptor(&manager);

    QDBusConnection::systemBus().registerObject(
        "/org/freedesktop/UDisks2", &manager);
    QDBusConnection::systemBus().registerObject(
        "/org/freedesktop/UDisks2/block_devices", new QObject());
    QDBusConnection::systemBus().registerObject(
        "/org/freedesktop/UDisks2/drives", new QObject());

    QThreadPool::globalInstance()->setExpiryTimeout(-1);
    QThreadPool::globalInstance()->setMaxThreadCount(4);

    initialGeomProbe();

    DevdThread devd;
    QObject::connect(&devd, &DevdThread::driveAdded, &manager, &ObjectManager::addDrive);
    QObject::connect(&devd, &DevdThread::driveRemoved, &manager, &ObjectManager::removeDrive);
    QObject::connect(&devd, &DevdThread::blockAdded, &manager, &ObjectManager::addBlock);
    QObject::connect(&devd, &DevdThread::blockRemoved, &manager, &ObjectManager::removeBlock);
    QObject::connect(&devd, &DevdThread::geomAdded, &manager, &ObjectManager::geomAdded);
    devd.start();

    app.exec();
}

#include "main.moc"
